@@ -141,7 +141,7 @@
  * The second lock (queue_lock) protects the various queues that are
  * used to implement the "fast" writes described above.
  *
- * If both locks are held at the same time thne the sources lock must
+ * If both locks are held at the same time then the sources lock must
  * have been acquired first.
  */
 
@@ -451,7 +451,7 @@ dconf_engine_read (DConfEngine *engine,
    *     means that no locks were found.  Non-zero means that a lock was
    *     found in the source with the index given by the variable.
    *
-   *  2. check the uncommited changes in the read_through list as the
+   *  2. check the uncommitted changes in the read_through list as the
    *     highest priority.  This is only done if we have a writable
    *     source and no locks were found.
    *
@@ -546,6 +546,51 @@ dconf_engine_read (DConfEngine *engine,
   return value;
 }
 
+GVariant *
+dconf_engine_read_user_value (DConfEngine *engine,
+                              GQueue      *read_through,
+                              const gchar *key)
+{
+  gboolean found_key = FALSE;
+  GVariant *value = NULL;
+
+  /* This is a simplified version of the above.  We get to ignore locks
+   * and system-level settings.
+   *
+   * NB: we may find "NULL", which is why we have a separate variable.
+   */
+
+  /* Ignore the queues if we don't have a writable database */
+  if (engine->n_sources == 0 || !engine->sources[0]->writable)
+    return NULL;
+
+  /* First check read-through */
+  if (read_through)
+    found_key = dconf_engine_find_key_in_queue (read_through, key, &value);
+
+  /* Next pending/in-flight */
+  if (!found_key)
+    {
+      dconf_engine_lock_queues (engine);
+
+      /* Check the pending queue first because those were submitted
+       * more recently.
+       */
+      found_key = dconf_engine_find_key_in_queue (&engine->pending, key, &value) ||
+                  dconf_engine_find_key_in_queue (&engine->in_flight, key, &value);
+
+      dconf_engine_unlock_queues (engine);
+    }
+
+  /* Finally, check the user database */
+  if (!found_key && engine->sources[0]->values)
+    value = gvdb_table_get_value (engine->sources[0]->values, key);
+
+  dconf_engine_release_sources (engine);
+
+  return value;
+}
+
 gchar **
 dconf_engine_list (DConfEngine *engine,
                    const gchar *dir,
@@ -878,7 +923,6 @@ dconf_engine_change_completed (DConfEngine  *engine,
                                const GError *error)
 {
   OutstandingChange *oc = handle;
-  DConfChangeset *expected;
 
   dconf_engine_lock_queues (engine);
 
@@ -888,9 +932,27 @@ dconf_engine_change_completed (DConfEngine  *engine,
    *
    * The reply we just received should therefore be at the head of
    * our 'in flight' queue.
+   *
+   * Due to https://bugs.freedesktop.org/show_bug.cgi?id=59780 it is
+   * possible that we receive an out-of-sequence error message, however,
+   * so only assume that messages are in-order for positive replies.
    */
-  expected = g_queue_pop_head (&engine->in_flight);
-  g_assert (expected && oc->change == expected);
+  if (reply)
+    {
+      DConfChangeset *expected;
+
+      expected = g_queue_pop_head (&engine->in_flight);
+      g_assert (expected && oc->change == expected);
+    }
+  else
+    {
+      gboolean found;
+
+      g_assert (error != NULL);
+
+      found = g_queue_remove (&engine->in_flight, oc->change);
+      g_assert (found);
+    }
 
   /* We just popped a change from the in-flight queue, possibly
    * making room for another to be added.  Check that.
@@ -920,7 +982,7 @@ dconf_engine_change_completed (DConfEngine  *engine,
 
   if (error)
     {
-      /* Some kind of unexpected failure occured while attempting to
+      /* Some kind of unexpected failure occurred while attempting to
        * commit the change.
        *
        * There's not much we can do here except to drop our local copy
@@ -1012,9 +1074,14 @@ dconf_engine_change_fast (DConfEngine     *engine,
 {
   GList *node;
 
+  if (dconf_changeset_is_empty (changeset))
+    return TRUE;
+
   if (!dconf_engine_changeset_changes_only_writable_keys (engine, changeset, error))
     return FALSE;
 
+  dconf_changeset_seal (changeset);
+
   /* Check for duplicates in the pending queue.
    *
    * Note: order doesn't really matter here since "similarity" is an
@@ -1074,9 +1141,19 @@ dconf_engine_change_sync (DConfEngine     *engine,
 {
   GVariant *reply;
 
+  if (dconf_changeset_is_empty (changeset))
+    {
+      if (tag)
+        *tag = g_strdup ("");
+
+      return TRUE;
+    }
+
   if (!dconf_engine_changeset_changes_only_writable_keys (engine, changeset, error))
     return FALSE;
 
+  dconf_changeset_seal (changeset);
+
   /* we know that we have at least one source because we checked writability */
   reply = dconf_engine_dbus_call_sync_func (engine->sources[0]->bus_type,
                                             engine->sources[0]->bus_name,
